# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/API/01_track.ipynb.

# %% auto 0
__all__ = ['Track']

# %% ../nbs/API/01_track.ipynb 5
from fastcore.basics import *

from bokeh.plotting import figure

from bokeh.models import (
    Quad,
    CustomJS,
    ColumnDataSource,
    NumeralTickFormatter,
    Range1d,
    HoverTool,
)

from .javascript import track_callback_code

import pandas as pd


try: #pyBigWig cannot be installed on Windows
    import pyBigWig
except ImportError:
    pyBigWig = None
    
import warnings



# %% ../nbs/API/01_track.ipynb 8
class Track:
    """ Track objects should only be created through GenomeBrowser.add_track """
    def __init__(self,
                 ylim: tuple = None, #limits of the y axis. If not specified, ylim will be set automatically with the max and min of the data plotted with Track.line, Track.scatter or Track.bar
                 height: int = 200, #size of the track
                 tools: str = "xwheel_zoom, ywheel_zoom, pan, box_zoom, save, reset", #comma separated list of Bokeh tools that can be used to navigate the plot
                 output_backend="webgl",
                 **kwargs,
                ):        
        self.height = height

        #ensuring that the active_scroll tool is part of the tools list 
        if "xwheel_zoom" not in [t.strip() for t in tools.split(',')]:
            tools+=", xwheel_zoom"

        self.fig = figure(tools=tools,
                          active_scroll="xwheel_zoom",
                          height=height,
                          y_axis_location="right", #this is required in order to keep a proper alignment with the sequence
                          output_backend=output_backend,
                          **kwargs)
        self.fig.xaxis[0].formatter = NumeralTickFormatter(format="0,0")
        self.track_loaded_data = None
        self.track_all_data = None
        self.loaded_range = None
        self.set_ylim(ylim)
            
    def set_ylim(self, ylim):
        self._ylim=ylim
        if self._ylim != None:
            self.fig.y_range=Range1d(ylim[0],ylim[1],
                                    bounds=ylim)
        
        


# %% ../nbs/API/01_track.ipynb 13
@patch
def _set_track_data_source(self:Track, data, pos, columns):
    columns=[c for c in columns if c] #some arguments can be None => remove them
    data=data.loc[(self.bounds[0] < data[pos]) & (data[pos] < self.bounds[1]),
                  [pos]+columns].copy()
    data=data.sort_values(pos)
    
    
    if len(data)>10**5:
        warnings.warn("You are trying to plot more than 10^5 glyphs, this might overflow your memory. \
        Consider using bounds or reducing the number of datapoints.")

    y=columns[0]
    if self._ylim == None:
        ymin = data[y].values.min()
        ymax = data[y].values.max()
        self._ylim = (ymin, ymax)
        self.fig.y_range=Range1d(ymin,ymax,
                                bounds=(ymin,ymax))
    
    self.all_data=ColumnDataSource(data)
    self.loaded_data=ColumnDataSource(
        data.loc[(self.loaded_range.data["start"][0] < data[pos]
                 ) & (
                 data[pos] < self.loaded_range.data["end"][0])]
    )
    
    xcb = CustomJS(
            args = {
                "x_range": self.fig.x_range,
                "pos": pos,
                "all_data":self.all_data,
                "loaded_data": self.loaded_data,
                "track_loaded_range":self.loaded_range,
            },
            code = track_callback_code
        )

    self.fig.x_range.js_on_change('start', xcb)
    tooltips=[(attr,f"@{attr}") for attr in set(columns)]
    self.fig.add_tools(HoverTool(tooltips=tooltips))


# %% ../nbs/API/01_track.ipynb 14
@patch
def line(self:Track,
         data: pd.DataFrame, #pandas DataFrame containing the data
         pos: str, #name of the column containing the positions along the genome
         y: str, #name of the column containing the data to be plotted on the y-axis
         hover_data: list = [], #list of column names to be shown when hovering over the data
         **kwargs #enables to pass keyword arguments used by the Bokeh function
        ):
    if type(hover_data)==str:
        hover_data = [hover_data]
        
    self._set_track_data_source(data, pos, columns=[y]+hover_data)
    self.fig.line(source=self.loaded_data, x=pos, y=y, **kwargs)


# %% ../nbs/API/01_track.ipynb 18
from bokeh.transform import factor_cmap

# %% ../nbs/API/01_track.ipynb 19
@patch
def scatter(self:Track,
         data: pd.DataFrame, #pandas DataFrame containing the data
         pos: str, #name of the column containing the positions along the genome
         y: str, #name of the column containing the data to be plotted on the y-axis
         factors: str = None, #name of a column of values to be used as factors
         hover_data: list = [], #list of additional column names to be shown when hovering over the data
         **kwargs, #enables to pass keyword arguments used by the Bokeh function
        ):
    if type(hover_data)==str:
        hover_data = [hover_data]
        
    self._set_track_data_source(data, pos, columns=[y,factors]+hover_data)
    
    if factors!=None:
        color=factor_cmap(factors,"Category10_10",tuple(set(data[factors].values)))
        
        self.fig.scatter(source=self.loaded_data, x=pos, y=y, color=color, legend_group=factors, **kwargs)
        
        self.fig.legend.title = factors
        self.fig.legend.location = "top_left"
    else:
        self.fig.scatter(source=self.loaded_data, x=pos, y=y, **kwargs)


# %% ../nbs/API/01_track.ipynb 25
@patch
def bar(self:Track,
         data: pd.DataFrame, #pandas DataFrame containing the data
         pos: str, #name of the column containing the positions along the genome
         y: str, #name of the column containing the data to be plotted on the y-axis
         factors: str = None, #name of a column of values to be used as factors
         hover_data: list = [], #list of additional column names to be shown when hovering over the data
         **kwargs, #enables to pass keyword arguments used by the Bokeh function
        ):
    
    if type(hover_data)==str:
        hover_data = [hover_data]
        
    self._set_track_data_source(data, pos, columns=[y,factors]+hover_data)
    
    if factors!=None:
        color=factor_cmap(factors,"Category10_3",tuple(set(data[factors].values)))
        
        self.fig.vbar(source=self.loaded_data, x=pos, top=y, color=color, legend_group=factors, **kwargs)

        self.fig.legend.location = "top_left"
        self.fig.legend.title = factors
    else:
        self.fig.vbar(source=self.loaded_data, x=pos, top=y, **kwargs)
        

# %% ../nbs/API/01_track.ipynb 29
@patch
def highlight(self:Track,
         data: pd.DataFrame, #pandas DataFrame containing the data
         left: str = "left", #name of the column containing the start positions of the regions
         right: str = "right", #name of the column containing the end positions of the regions
         color: str = "color", #color of the regions
         alpha: str = 0.2, #transparency
         hover_data: list = [], #list of additional column names to be shown when hovering over the data
         **kwargs, #enables to pass keyword arguments used by the Bokeh function
        ):
    
    if type(hover_data)==str:
        hover_data = [hover_data]

    if color not in data.columns:
        data["color"]='green'

    data["alpha"]=alpha

    highlight_source = ColumnDataSource(data[[left,right,"color","alpha"]+hover_data])

    if self._ylim==None:
        warnings.warn("When adding highlights to a track, ylim needs to be defined. \
                      You can eigher set ylim manually when creating the track, or plot data using Track.line, Track.scatter or Track.bar before adding the highlight.")
    
    r=Quad(left=left, right=right,
           bottom=self._ylim[0],
           top=self._ylim[1],
           fill_color="color",
           fill_alpha="alpha",
           line_alpha=0,
           **kwargs)

    renderer= self.fig.add_glyph(highlight_source, r)
    tooltips=[(f"{left} - {right}",f"@{left} - @{right}")]+[(f"{attr}",f"@{attr}") for attr in hover_data]
    self.fig.add_tools(HoverTool(renderers=[renderer],
                                        tooltips=tooltips))
